use super::*;

/// Failure yielded by [`Origin`].
#[derive(Debug)]
pub enum ResolutionError<L, P> {
    Lookup(L),
    Parse(P),
}

/// See [`ResolutionResult`].
pub type ResolutionFailure<'a, Ctx, A> = ResolutionError<
    <Ctx as Context>::LookupError<'a>,
    <<A as Mentionable<'a, Ctx>>::Fctr as Factory<'a, Ctx>>::ParseError,
>;

/// Result yielded by [`Origin`].
pub type ResolutionResult<'a, Ctx, A> = Result<Rc<A>, ResolutionFailure<'a, Ctx, A>>;

/// Wrapped result returned by [`Origin`].
pub type Resolution<'a, Ctx, A> = Wrapped<'a, Ctx, ResolutionResult<'a, Ctx, A>>;

/// Short-hand for the type of values returned by [`Resolver::resolve`].
pub type HashResolution<'a, Ctx> = Wrapped<
    'a,
    Ctx,
    Result<(Vec<u8>, Rc<dyn Resolver<'a, Ctx>>), <Ctx as Context>::LookupError<'a>>,
>;

/// Value accepted by [`Resolver::resolve`]. Includes index to make it order-sensitive.
#[derive(Clone, Copy, Debug)]
pub struct Address {
    pub point: Hash,
    /// Index of the point in the [`Mentionable::points()`].
    pub index: usize,
}

/// Trait representing the "rainbow table" behaviour.
pub trait Resolver<'a, Ctx: 'a + Context>: 'a {
    /// Successfully returned value should be the inverse of the point passed
    /// with topology header ([`Mentionable::topology()`]) omitted.
    fn resolve(self: Rc<Self>, address: Address) -> HashResolution<'a, Ctx>;
}

/// Preferred way to parse [Point]s off of a [Serializer].
pub struct Addresses {
    current: usize,
}

impl Addresses {
    /// Read the next [Address].
    pub fn next<'a>(
        &mut self,
        deserializer: &'a mut dyn Deserializer,
    ) -> Result<Address, &'a [u8]> {
        let point = deserializer.read_n_const::<HASH_SIZE>()?;
        let address = Address {
            point,
            index: self.current,
        };
        self.current += 1;
        Ok(address)
    }

    /// Read the next [Point].
    pub fn next_point<'a, 'b, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
        &mut self,
        deserializer: &'b mut dyn Deserializer,
        resolver: Rc<dyn Resolver<'a, Ctx>>,
        factory: A::Fctr,
    ) -> Result<Point<'a, Ctx, A>, &'b [u8]> {
        Ok(Point::from_address(
            self.next(deserializer)?,
            factory,
            resolver,
        ))
    }

    /// Start reading the [Address]es.
    fn start() -> Self {
        Addresses { current: 0 }
    }
}

fn _parse_slice<'a, Ctx: 'a + Context, A: Mentionable<'a, Ctx>>(
    factory: &A::Fctr,
    slice: &[u8],
    resolver: Rc<dyn Resolver<'a, Ctx>>,
) -> ParseResult<'a, Ctx, A::Fctr> {
    let mut deserializer = SliceDeserializer::from(slice);
    let mentionable = factory.deserialize(&mut deserializer, resolver, &mut Addresses::start())?;
    let tail = deserializer.read_all();
    if tail.is_empty() {
        Ok(mentionable)
    } else {
        Err(factory.unexpected_tail(tail))
    }
}

impl<'a, Ctx: 'a + Context, F: Factory<'a, Ctx>> ExtFactory<'a, Ctx> for F {
    fn parse_slice(
        &self,
        slice: &[u8],
        resolver: Rc<dyn Resolver<'a, Ctx>>,
    ) -> ParseResult<'a, Ctx, Self> {
        _parse_slice::<Ctx, F::Mtbl>(self, slice, resolver)
    }
}
